// Copyright (c) 2005-2010 StreamBase Systems, Inc. All rights reserved.
package com.streambase.contrib.eclipse.connectivity.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.LinkedBlockingQueue;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.ListenerList;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;

import com.streambase.contrib.eclipse.connectivity.ISBConnection;
import com.streambase.contrib.eclipse.connectivity.ISBDequeueListener;
import com.streambase.contrib.eclipse.connectivity.IStatusListener;
import com.streambase.contrib.eclipse.connectivity.SBConnectivityPlugin;
import com.streambase.sb.EntityType;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamProperties;
import com.streambase.sb.Tuple;
import com.streambase.sb.client.DequeueResult;
import com.streambase.sb.client.StreamBaseAdminClient;
import com.streambase.sb.client.StreamBaseClient;
import com.streambase.sb.client.StreamBaseURI;

/**
 * Internal class representing a connection to a StreamBase Server. The only instance
 * of this class (as an {@link ISBConnection}) should be obtained via {@link SBConnectivityPlugin#getSBConnection()}.
 */
public class SBConnection implements ISBConnection {

	public class DequeueThread extends Thread {
		private StreamBaseException deqException;
		private StreamBaseClient dequeueClient;

		public DequeueThread(String[] dequeueStreams) throws StreamBaseException {
			this.dequeueClient = new StreamBaseClient(targetURI);
			if (dequeueStreams != null) {
                for (String stream : dequeueStreams) {
                    try {
                        this.dequeueClient.subscribe(stream);
                    } catch (StreamBaseException e) {
                        // ignore this bad stream, and keep going with the other streams
                    	SBConnectivityPluginActivator.getDefault().getLog().log((new Status(IStatus.ERROR, SBConnectivityPluginActivator.PLUGIN_ID, "StreamBaseException", e)));
                    }
                }
			}
		}

		public void die() {
			try {
				// "wake up" our dequeue thread
				dequeueClient.close();
			} catch (StreamBaseException e) {
				// ignore
			}
		}
		public void reportStatus(MultiStatus status) {
			if (deqException != null) {
				status.add(new Status(IStatus.ERROR, SBConnectivityPluginActivator.PLUGIN_ID, "Dequeuer exited with exception", deqException));
			} else if (!isAlive()) {
				status.add(new Status(IStatus.WARNING, SBConnectivityPluginActivator.PLUGIN_ID, "Dequeuer has exited"));
			} else {
				status.add(new Status(IStatus.OK, SBConnectivityPluginActivator.PLUGIN_ID, null));
			}
		}

		@Override
		public void run() {
			while (true) {
				try {
					DequeueResult result = dequeueClient.dequeue();
					if (result == null || result.getStatus() != DequeueResult.GOOD) { // connection closed or something
						if (!die) {
							throw new StreamBaseException("dequeue() returned no good result");
						}
						return;
					}
					
					// Give the DequeueResult to registered listeners:
					Object[] listeners = dequeueListeners.getListeners();
					for (int i = 0; i < listeners.length; ++i) {
						((ISBDequeueListener) listeners[i]).tuplesDequeued(result);
					}
					
					// Find TupleHandlers for the stream of the DequeueResult.
					// Then give each tuple of the DequeueResult to the handler
					// method of each TupleHandler:
					String streamName = result.getStreamProperties().getQualifiedName();
					List<SBTupleHandler> handlers = SBDequeuedTupleDispatcher.getDefault().getTupleHandlers (streamName);
					if (handlers != null) {
						for (Tuple tuple : result) {
							for (SBTupleHandler handler : handlers) {
								try {
									handler.invoke (tuple, streamName);
								}
								catch (Exception e) {
									SBConnectivityPluginActivator.logError (
											"Exception thrown by dequeued Tuple handler, \"" + handler + "\"", e);
								}
							}
						}
					}
				}
				catch (final StreamBaseException e) {
					// failure to dequeue; this thread goes bye-bye after noting this
					deqException = e;
					// update the status
					updateStatus();
					return;
				}
			}
		}
	}
	
	private class EnqElement {
    	String stream;
    	Collection<Tuple> tuples;
    	public EnqElement(String stream, Collection<Tuple> tuples) {
    		this.stream = stream;
    		this.tuples = tuples;
    	}
    }

    public class EnqueueThread extends Thread{
    	private StreamBaseException enqException;
    	private StreamBaseClient enqueueClient;
    
    	public EnqueueThread() throws StreamBaseException {
    		this.enqueueClient = new StreamBaseClient(targetURI);
    	}
    
    	public void reportStatus(MultiStatus status) {
    		if (enqException != null) {
    			status.add(new Status(IStatus.ERROR, SBConnectivityPluginActivator.PLUGIN_ID, "Enqueuer exited with exception", enqException));
    		} else if (!isAlive()) {
    			status.add(new Status(IStatus.WARNING, SBConnectivityPluginActivator.PLUGIN_ID, "Enqueuer has exited"));
    		} else {
    			status.add(new Status(IStatus.OK, SBConnectivityPluginActivator.PLUGIN_ID, null));
    		}
    	}
    
    	@Override
    	public void run() {
    		while (true) {
    			try {
    				EnqElement enqElement = enqElements.take();
    				enqueueClient.enqueue(enqElement.stream, enqElement.tuples);
    			} catch (InterruptedException e) {
    				// was interrupted while waiting for an element; check if we are going down
    				if (die) {
    					try {
    						enqueueClient.close();
    					} catch (StreamBaseException e1) {
    						// ignore close exception on die
    					}
    					return;
    				}
    			} catch (final StreamBaseException e) {
    				// failure to dequeue; this thread goes bye-bye after noting this
    				enqException = e;
    				updateStatus();
    				return;
    			}
    		}
    	}
    }
    
    private static SBConnection instance;
    
    public static SBConnection getInstance() {
        return instance;
    }

    private StreamBaseURI targetURI;
    
    /**
     * Returns the StreamBaseURI that this connection is for
     * @TestingUse
     * @return the StreamBaseURI that this connection is for
     */
	public StreamBaseURI getTargetURI() {
		return targetURI;
	}

	public StreamBaseClient getDequeueClient() {
		return deqThread.dequeueClient;
	}

	public StreamBaseClient getEnqueueClient() {
		return enqThread.enqueueClient;
	}

	private DequeueThread deqThread;
	private ListenerList dequeueListeners = new ListenerList(); // of ISBDequeueListener
	private String[] initialDequeueStreams;
	
	public String[] getInitialDequeueStreams() {
		return initialDequeueStreams;
	}

	private ISBDequeueListener[] initialDequeueListeners;
	   
    private LinkedBlockingQueue<EnqElement> enqElements = new LinkedBlockingQueue<EnqElement>();
    private Map<String, Schema> inputStreamSchemas = new HashMap<String, Schema>();
    private Map<String, Schema> outputStreamSchemas = new HashMap<String, Schema>();
    private EnqueueThread enqThread;
	private boolean die = false;
	
	   private ListenerList statusListeners = new ListenerList();
	
	/* package */ SBConnection(String[] initialDequeueStreams, ISBDequeueListener[] initialDequeueListeners) {
		this.initialDequeueStreams = initialDequeueStreams;
		this.initialDequeueListeners = initialDequeueListeners;
	}

	/* (non-Javadoc)
	 * @see com.streambase.sb.rcp.connectivity.internal.ISBConnection#connectTo(java.lang.String)
	 */
	public void connectTo(StreamBaseURI server) throws StreamBaseException, IllegalStateException {
		if (isConnected()) {
			throw new IllegalStateException("already connected");
		}
		this.targetURI = server;
		// immediate connectivity check
		{
			StreamBaseAdminClient admin = new StreamBaseAdminClient(targetURI);
			admin.status();
			admin.close();
		}
		die = false;
		if (initialDequeueListeners != null) { // initial dequeue listeners registered now
			for (ISBDequeueListener listener : initialDequeueListeners) {
				dequeueListeners.add(listener);
			}
		}
		try {
			// create enqueue thread (connect-on-start)
		    enqThread = new EnqueueThread();

	        enqThread.start();
            collectStreamSchemas();
			// create dequeue thread (connects immediately at constructor)
			deqThread = new DequeueThread(initialDequeueStreams);
			deqThread.start();
			updateStatus();
		} catch (Throwable e) { // on any failure, lose all listeners we tried to put
			dequeueListeners.clear();
		}
	}

	private boolean isConnected() {
		return deqThread != null && deqThread.isAlive();
	}
	
	/* (non-Javadoc)
	 * @see com.streambase.sb.rcp.connectivity.ISBConnection#disconnect()
	 */
	public void disconnect() {
		die = true;
		// kill the dequeue thread
		if (deqThread != null) {
			deqThread.interrupt();
			try {
				deqThread.join(2500);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		// kill the enqueue thread
		if (enqThread != null) {
            enqThread.interrupt();
            try {
                enqThread.join(2500);

            } catch (InterruptedException e) {
                // ignore

            }
		}
		    
	}

	
	/* (non-Javadoc)
	 * @see com.streambase.sb.rcp.connectivity.ISBConnection#addDequeueListener(com.streambase.sb.rcp.connectivity.ISBDequeueListener)
	 */
	public void addDequeueListener(ISBDequeueListener l) {
		dequeueListeners.add(l);
	}
	
	/* (non-Javadoc)
	 * @see com.streambase.sb.rcp.connectivity.ISBConnection#removeDequeueListener(com.streambase.sb.rcp.connectivity.ISBDequeueListener)
	 */
	public void removeDequeueListener(ISBDequeueListener l) {
		dequeueListeners.remove(l);
	}
	
	/**
	 * Enqueue a tuple onto a stream
	 * @param stream The stream to enqueue on
	 * @param tuple The tuple to enqueue
	 * @throws StreamBaseException thrown if there is a problem enqueueing the tuple
	 */
    public void enqueue (String stream, Tuple tuple) throws StreamBaseException {
		List<Tuple> tuples = new ArrayList<Tuple>(1);
		tuples.add (tuple);
		enqueue (stream, tuples);
    }
	
	/**
     * Enqueue the given tuples asynchronously. May be called from any thread.
     * Caller must not modify the passed in tuple collection! 
     * @param stream the target stream name
     * @param tuples tuples to enqueue
     * @throws StreamBaseException if there is no enqueue connection to the server, or if the local buffer (very, very large!) ever gets full
     */
    public void enqueue(String stream, Collection<Tuple> tuples) throws StreamBaseException {
        if (!enqThread.isAlive()) throw new StreamBaseException("No enqueue connection to server");
        try {
            enqElements.add(new EnqElement(stream, tuples));
        } catch (IllegalStateException e) {
            throw new StreamBaseException("cannot accept tuples for enqueue; local enqueue buffer full");
        }
    }
	   
	/**
	 * Returns the Schema for the passed input stream name.
	 * @param stream Module qualified stream name
	 * @return the Schema for the passed stream. null if not found.
	 */
    public Schema getInputStreamSchemaFromName(String stream) {
        return inputStreamSchemas.get(stream);
    }
	
    

    public Schema getOutputStreamSchemaFromName(String stream){
    	return outputStreamSchemas.get(stream);
    }
    
    /**
     * Loads the input and output schemas from the StreamBase app running at the current {@link #targetURI}
     * @throws StreamBaseException
     */
    private void collectStreamSchemas() throws StreamBaseException {
    	StreamBaseClient client = new StreamBaseClient(targetURI);

    	Set<StreamProperties> inputProps =  client.getAllStreamProperties(EntityType.INPUT_STREAMS);
    	for (StreamProperties streamProperties : inputProps) {
    		inputStreamSchemas.put(streamProperties.getModulePath().toString(),streamProperties.getSchema());
    	}
    	
    	Set<StreamProperties> outputProps =  client.getAllStreamProperties(EntityType.OUTPUT_STREAMS);
    	for (StreamProperties streamProperties : outputProps) {
    		outputStreamSchemas.put(streamProperties.getModulePath().toString(),streamProperties.getSchema());
    	}
        client.close();
    }
    
    public void addStatusListener(IStatusListener l) {
        statusListeners.add(l);
    }
    
    /**
     * @return a MultiStatus whose first child represents the dequeuer state, the second child represents the enqueuer state
     */
    public IStatus getStatus() {
        MultiStatus status = new MultiStatus(SBConnectivityPluginActivator.PLUGIN_ID, IStatus.OK, "Server Connection Status", null);
        if (deqThread != null) {
            deqThread.reportStatus(status);
        }
        if (enqThread != null) {
            enqThread.reportStatus(status);
        }
        return status;
    }
    
    public void removeStatusListener(IStatusListener l) {
        statusListeners.remove(l);
    }

    public void updateStatus() {
        IStatus status = getStatus();
        Object[] listeners = statusListeners.getListeners();
        for (int i = 0; i < listeners.length; ++i) {
            ((IStatusListener) listeners[i]).connectionStatusChanged(status);
        }
    }
    
}
