// Copyright (c) 2005-2010 StreamBase Systems, Inc. All rights reserved.
package com.streambase.contrib.eclipse.connectivity.internal;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;
import org.osgi.framework.BundleContext;

import com.streambase.contrib.eclipse.connectivity.ISBDequeueListener;


/**
 * Internal activator class controlling plug-in life cycle.
 * This plug-in activates on any class load. During activation,
 * this plug-in looks for any (single) registered connection extension, and
 * immediately attempts a connection job to the remote server. Regardless
 * of the connection status, this plug-in always starts up (failure to connect
 * is a status of the {@link SBConnection} instance, not a bundle problem).
 */
public class SBConnectivityPluginActivator extends Plugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "com.streambase.contrib.eclipse.connectivity";

	// The shared instance
	private static SBConnectivityPluginActivator plugin;
	
	private SBConnection connection;
	
	/**
	 * The public constructor for the Platform. Clients should not instantiate.
	 */
	public SBConnectivityPluginActivator() {
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		plugin = this;
		
		// This plugin supports two different kinds of extensions for handling
		// dequeued Tuples:

		// (1) Process all tuple consumer methods declared via out "dequeuedTupleConsumers" extension point
		Set<String> initialStreams = SBDequeuedTupleDispatcher.getDefault().processDequeuedTupleConsumerExtensions();
		
		// (2) Process all initial dequeuers declared via our "initialConnectionParameters" extension point
		List<ISBDequeueListener> initialDequeuers = new ArrayList<ISBDequeueListener>();
		IExtension[] extensions = Platform.getExtensionRegistry().getExtensionPoint(PLUGIN_ID, "initialConnectionParameters").getExtensions();
		for (IExtension extension : extensions) {
			IConfigurationElement[] configurationElements = extension.getConfigurationElements();
			for (IConfigurationElement configurationElement : configurationElements) {
				if (configurationElement.getName().equals("dequeuers")) {
					// initial streams
					String streams = configurationElement.getAttribute("streams");
					if (streams != null) {
						initialStreams.addAll(Arrays.asList(streams.split(",")));
					}
					// dequeue listeners
					IConfigurationElement[] deqListenerElements = configurationElement.getChildren("initialDequeueListener");
					for (IConfigurationElement deqListenerElement : deqListenerElements) {
						try {
							initialDequeuers.add((ISBDequeueListener) deqListenerElement.createExecutableExtension("class"));
						} catch (CoreException e) {
							// failed to create instance; log
							logError("Unable to create dequeue listener from extension", e);
						}
					}
				}
			}
		}
		// create our singleton connection
		connection = new SBConnection(
				initialStreams.toArray(new String[initialStreams.size()]),
				initialDequeuers.toArray(new ISBDequeueListener[initialDequeuers.size()])); 
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
		plugin = null;
		connection.disconnect();
		super.stop(context);
	}

	/**
	 * Returns the shared instance
	 *
	 * @return the shared instance
	 */
	public static SBConnectivityPluginActivator getDefault() {
		return plugin;
	}

    /**
     * Logs the given error message to the platform log.
     * 
     * @param message A high level UI message describing when the problem happened.
     */
    public static void logError(String message) {
        getDefault().getLog().log (
        		new Status(IStatus.ERROR, PLUGIN_ID, message));
    }
    
    /**
	 * Logs the given message and throwable to the platform log.
	 * 
	 * @param message A high level UI message describing when the problem happened.
	 * @param t The throwable from where the problem actually occurred.
	 */
    public static void logError (String message, Throwable t) {
        IStatus status = new Status(IStatus.ERROR, PLUGIN_ID, message, t);
        log(message, status);
    }
    
    /**
	 * Logs the given throwable to the platform log.
	 * 
	 * @param t The throwable from where the problem actually occurred.
	 */
    public static void logError (Throwable t) {
    	logError (t.getClass().getName(), t);
    }
    
    /**
     * Logs the given message and status to the platform log. (If
     * the message is non-null, it is logged before the status as an individual
     * error log entry).
     * 
     * This convenience method is for internal use.
     * 
     * @param message
     *            A high level UI message describing when the problem happened.
     *            May be <code>null</code>.
     * @param status
     *            The status describing the problem. Must not be null.
     */
    public static void log(String message, IStatus status) {
        if (message != null) {
            getDefault().getLog().log(
                    new Status(IStatus.ERROR, PLUGIN_ID, message, null));
        }

        getDefault().getLog().log(status);
    }
    
    /**
     * Log the status to the platform log.
     * @param status  The status describing the problem. Must not be null.
     */
    public static void log(IStatus status) {
        getDefault().getLog().log(status);
    }
    
	public SBConnection getSBConnection() {
		return connection;
	}
}
