// Copyright (c) 2005-2010 StreamBase Systems, Inc. All rights reserved.
package com.streambase.contrib.eclipse.connectivity.internal;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.Platform;

import com.streambase.sb.Tuple;


/**
 * This singleton manages "dequeuedTupleConsumers" Eclipse extensions,
 * which let users of this plugin declare Java methods to handle
 * Tuples dequeued from specific output streams.
 */
public class SBDequeuedTupleDispatcher {
	
	// -- static members ---
	
	protected static SBDequeuedTupleDispatcher sSingleton = new SBDequeuedTupleDispatcher();

	
	// -- members ---
	
	/**
	 * Maps fully-qualified stream names to handlers of Tuples dequeued
	 * from those streams.  Each handler has a cached reference to the
	 * singleton Object that should be given each Tuple, along with a
	 * Method object (belonging to the singleton's class) which is invoked
	 * for each dequeued Tuple.
	 */
	protected HashMap <String,ArrayList<SBTupleHandler>> fStreamToHandlersMap;
	
	
	// --- constructors ---
	
	private SBDequeuedTupleDispatcher() {
		// "private" because there's just the one singleton
		fStreamToHandlersMap = new HashMap <String,ArrayList<SBTupleHandler>> ();
	}
	
	
	// --- methods ---
	
	public static SBDequeuedTupleDispatcher getDefault() {
		return sSingleton;
	}

	/**
	 * Populates the map of stream names to tuple handlers
	 * and returns a set of all the stream names.
	 */
	public HashSet<String> processDequeuedTupleConsumerExtensions() {
		HashSet<String> streamsToSubscribeTo = new HashSet<String>();
		IExtension[] extensions = Platform.getExtensionRegistry().getExtensionPoint (
				SBConnectivityPluginActivator.PLUGIN_ID, "dequeuedTupleConsumers").getExtensions();
		for (IExtension extension : extensions) {
			IConfigurationElement[] elements = extension.getConfigurationElements();
			for (IConfigurationElement element : elements) {
				if (element.getName().equals("dequeuedTupleConsumer")) {
					Object classInstance = null;
					try {
						classInstance = element.createExecutableExtension("class");
					}
					catch (CoreException e) {
						SBConnectivityPluginActivator.logError (
								"\"dequeuedTupleConsumer\" class not found", e);
					}
					String singletonMethodName = element.getAttribute ("nameOfGetSingletonMethod");
					Object singleton = getSingleton (classInstance, singletonMethodName);
					if (singleton == null) {
						continue;
					}
					IConfigurationElement[] streamAndMethodPairs = element.getChildren("streamAndTupleConsumerMethod");
					for (IConfigurationElement streamAndMethodPair : streamAndMethodPairs) {
						String streamName = streamAndMethodPair.getAttribute ("streamName");
						String methodName = streamAndMethodPair.getAttribute ("methodName");
						Method consumeTupleMethod = getConsumeTupleMethod (singleton, methodName);
						if (consumeTupleMethod == null) {
							continue;
						}
						registerTupleHandlerMethod (streamName, singleton, consumeTupleMethod);
						streamsToSubscribeTo.add (streamName);
					}
				}
			}
		}
		return streamsToSubscribeTo;
	}
	
	/**
	 * Returns the TupleHandlers for tuples dequeued from the stream of
	 * the given name.
	 */
	public List<SBTupleHandler> getTupleHandlers (String streamName) {
		return fStreamToHandlersMap.get (streamName);
	}

	/**
	 * Adds a TupleHandler for the given "handlerSingleton" and
	 * "consumeTupleMethod" to the list of handlers associated
	 * with the stream of the given "streamName".
	 */
	private void registerTupleHandlerMethod (String streamName, Object handlerSingleton, Method consumeTupleMethod)
	{
		ArrayList<SBTupleHandler> handlers = fStreamToHandlersMap.get (streamName);
		if (handlers == null) {
			handlers = new ArrayList<SBTupleHandler>();
			fStreamToHandlersMap.put (streamName, handlers);
		}
		SBTupleHandler handler = new SBTupleHandler (handlerSingleton, consumeTupleMethod);
		handlers.add (handler);
	}

	/**
	 * Uses reflection to call a method with the given "singletonMethodName"
	 * on the given "classInstance", returning the result.
	 */
	private Object getSingleton (Object classInstance, String singletonMethodName)
	{
		Class<?> clazz = classInstance.getClass();
		
		// Get the static method which returns the class' singleton:
		Method getSingletonMethod = null;
		try {
			getSingletonMethod = clazz.getMethod (singletonMethodName, new Class<?>[0]);
		}
		catch (NoSuchMethodException e) {
			SBConnectivityPluginActivator.logError (
					"\"dequeuedTupleConsumer\" class' method, " +
					"\"" + clazz.getName() + "." + singletonMethodName + "\", not found", e);
			return null;
		}
		
		// Call the method which returns the class' singleton:
		Object result = null;
		try {
			result = getSingletonMethod.invoke (classInstance);
		}
		catch (Exception e) {
			SBConnectivityPluginActivator.logError (
					"Exception trying to call \"dequeuedTupleConsumer\" class' method, " +
					"\"" + clazz.getName() + "." + singletonMethodName + "()\"", e);
			return null;
		}
		
		// Complain if the method returned null:
		if (result == null) {
			SBConnectivityPluginActivator.logError (
					"Call to \"dequeuedTupleConsumer\" class' method, " +
					"\"" + clazz.getName() + "." + singletonMethodName + "()\", returned null");
			return null;
		}
		return result;
	}

	/**
	 * Uses reflection to return a method with the given "methodName"
	 * of the class of the given "singleton" Object.
	 */
	private Method getConsumeTupleMethod (Object singleton, String methodName)
	{
		Class<?> clazz = singleton.getClass();
		
		try {
			return clazz.getMethod (methodName, Tuple.class, String.class);
		}
		catch (NoSuchMethodException e) {
			SBConnectivityPluginActivator.logError (
					"No such \"dequeuedTupleConsumer\" (public) method, " +
					"\"" + clazz.getName() + "." + methodName + "\"");
			return null;
		}
	}
}
